#include "mex.h"
#include "fdlda-estimate.h"




/*
 * perform inference on a document and update sufficient statistics
 *
 */

double doc_e_step(document* doc, double* gamma, double* phi,
                  lda_model* model, lda_suffstats* ss)
{
    double likelihood;
    int n, k;

    // posterior inference

    likelihood = lda_inference(doc, model, gamma, phi);

    // update sufficient statistics

    double gamma_sum = 0;
    for (k = 0; k < model->num_topics; k++)
    {
        gamma_sum += gamma[k];
        ss->alpha_suffstats += digamma(gamma[k]);
    }
    ss->alpha_suffstats -= model->num_topics * digamma(gamma_sum);

    for (n = 0; n < doc->length; n++)
    {
        for (k = 0; k < model->num_topics; k++)
        {
            ss->class_word[k][doc->words[n]] += doc->counts[n]*phi[k]+LAP;
            ss->class_total[k] += doc->counts[n]*phi[k]+LAP;
        }
    }

    ss->num_docs = ss->num_docs + 1;

    return(likelihood);
}


/*
 * writes the word assignments line for a document to a file
 *
 */

void write_word_assignment(FILE* f, document* doc, double** phi, lda_model* model)
{
    int n;

    fprintf(f, "%03d", doc->length);
    for (n = 0; n < doc->length; n++)
    {
        fprintf(f, " %04d:%02d",
                doc->words[n], argmax(phi[n], model->num_topics));
    }
    fprintf(f, "\n");
    fflush(f);
}


/*
 * saves the gamma parameters of the current dataset
 *
 */

void save_gamma(char* filename, double** gamma, int num_docs, int num_topics)
{
    FILE* fileptr;
    int d, k;
    fileptr = fopen(filename, "w");

    for (d = 0; d < num_docs; d++)
    {
	fprintf(fileptr, "%5.10f", gamma[d][0]);
	for (k = 1; k < num_topics; k++)
	{
	    fprintf(fileptr, " %5.10f", gamma[d][k]);
	}
	fprintf(fileptr, "\n");
    }
    fclose(fileptr);
}


/*
 * run_em
 *
 */

void run_em(char* start, char* directory, corpus* corpus)
{

    int d, n;
    lda_model *model = NULL;
    double **var_gamma, **phi;

    // allocate variational parameters

    var_gamma = malloc(sizeof(double*)*(corpus->num_docs));
    for (d = 0; d < corpus->num_docs; d++)
	var_gamma[d] = malloc(sizeof(double) * NTOPICS);

    int max_length = max_corpus_length(corpus);
    phi = malloc(sizeof(double*)*max_length);
    for (n = 0; n < max_length; n++)
	phi[n] = malloc(sizeof(double) * NTOPICS);

    // initialize model

    char filename[100];

    lda_suffstats* ss = NULL;
    if (strcmp(start, "seeded")==0)
    {
        model = new_lda_model(corpus->num_terms, NTOPICS);
        ss = new_lda_suffstats(model);
        corpus_initialize_ss(ss, model, corpus);
        lda_mle(model, ss, 0);
        model->alpha = INITIAL_ALPHA;
    }
    else if (strcmp(start, "random")==0)
    {
        model = new_lda_model(corpus->num_terms, NTOPICS);
        ss = new_lda_suffstats(model);
        random_initialize_ss(ss, model);
        lda_mle(model, ss, 0);
        model->alpha = INITIAL_ALPHA;
    }
    else
    {
        model = load_lda_model(start);
        ss = new_lda_suffstats(model);
    }

    sprintf(filename,"%s/000",directory);
    save_lda_model(model, filename);

    // run expectation maximization

    int i = 0;
   

    double likelihood, likelihood_old = 0, converged = 1;
    sprintf(filename, "%s/likelihood.dat", directory);
    FILE* likelihood_file = fopen(filename, "w");

    while (((converged < 0) || (converged > EM_CONVERGED) || (i <= 2)) && (i <= EM_MAX_ITER))
    {
        i++; printf("**** em iteration %d ****\n", i);
        likelihood = 0;
        zero_initialize_ss(ss, model);


      

        // e-step

        for (d = 0; d < corpus->num_docs; d++)
        {
            if ((d % 1000) == 0) printf("document %d\n",d);
            likelihood += doc_e_step(&(corpus->docs[d]),
                                     var_gamma[d],
                                     phi,
                                     model,
                                     ss);
 
         
        }
       

        // m-step

        lda_mle(model, ss, ESTIMATE_ALPHA);

        // check for convergence

        converged = (likelihood_old - likelihood) / (likelihood_old);
        if (converged < 0) VAR_MAX_ITER = VAR_MAX_ITER * 2;
        likelihood_old = likelihood;

        // output model and likelihood

        fprintf(likelihood_file, "%10.10f\t%5.5e\n", likelihood, converged);
        fflush(likelihood_file);
        if ((i % LAG) == 0)
        {
            sprintf(filename,"%s/%03d",directory, i);
            save_lda_model(model, filename);
            sprintf(filename,"%s/%03d.gamma",directory, i);
            save_gamma(filename, var_gamma, corpus->num_docs, model->num_topics);
        }
    }

    // output the final model

    sprintf(filename,"%s/final",directory);
    save_lda_model(model, filename);
    sprintf(filename,"%s/final.gamma",directory);
    save_gamma(filename, var_gamma, corpus->num_docs, model->num_topics);

    // output the word assignments (for visualization)

    sprintf(filename, "%s/word-assignments.dat", directory);
    FILE* w_asgn_file = fopen(filename, "w");
    for (d = 0; d < corpus->num_docs; d++)
    {
        if ((d % 100) == 0) printf("final e step document %d\n",d);
        likelihood += lda_inference(&(corpus->docs[d]), model, var_gamma[d], phi);
        write_word_assignment(w_asgn_file, &(corpus->docs[d]), phi, model);
    }
    fclose(w_asgn_file);
    fclose(likelihood_file);
}


/*
 * read settings.
 *
 */

void read_settings(char* filename)
{
    FILE* fileptr;
    char alpha_action[100];
    fileptr = fopen(filename, "r");
    fscanf(fileptr, "var max iter %d\n", &VAR_MAX_ITER);
    fscanf(fileptr, "var convergence %f\n", &VAR_CONVERGED);
    fscanf(fileptr, "em max iter %d\n", &EM_MAX_ITER);
    fscanf(fileptr, "em convergence %f\n", &EM_CONVERGED);
    fscanf(fileptr, "alpha %s", alpha_action);
    if (strcmp(alpha_action, "fixed")==0)
    {
	ESTIMATE_ALPHA = 0;
    }
    else
    {
	ESTIMATE_ALPHA = 1;
    }
    fclose(fileptr);
}


/*
 * inference only
 *
 */

void infer(char* model_root, char* save, corpus* corpus)
{
    FILE* fileptr;
    char filename[100];
    int i, d, n;
    lda_model *model;
    double **var_gamma, likelihood, **phi;
    document* doc;

    model = load_lda_model(model_root);
    var_gamma = malloc(sizeof(double*)*(corpus->num_docs));
    for (i = 0; i < corpus->num_docs; i++)
	var_gamma[i] = malloc(sizeof(double)*model->num_topics);
    sprintf(filename, "%s-lda-lhood.dat", save);
    fileptr = fopen(filename, "w");
    for (d = 0; d < corpus->num_docs; d++)
    {
	if (((d % 100) == 0) && (d>0)) printf("document %d\n",d);

	doc = &(corpus->docs[d]);
	phi = (double**) malloc(sizeof(double*) * doc->length);
	for (n = 0; n < doc->length; n++)
	    phi[n] = (double*) malloc(sizeof(double) * model->num_topics);
	likelihood = lda_inference(doc, model, var_gamma[d], phi);

	fprintf(fileptr, "%5.5f\n", likelihood);
    }
    fclose(fileptr);
    sprintf(filename, "%s-gamma.dat", save);
    save_gamma(filename, var_gamma, corpus->num_docs, model->num_topics);
}







/*
 * run_em
 *
 */

lda_model* run_em_mat(int start, corpus* corpus,double ***var_g,double ***phii,double **nu)
{

    int d, n;
    double **phi;
    lda_model *model = NULL;
    double **var_gamma;

    // allocate variational parameters

    var_gamma = malloc(sizeof(double*)*(corpus->num_docs));
    for (d = 0; d < corpus->num_docs; d++)
	var_gamma[d] = malloc(sizeof(double) * NTOPICS);

    int max_length = max_corpus_length(corpus);
    //phi = malloc(sizeof(double*)*max_length);
    phi = malloc(sizeof(double*)*(corpus->num_docs));
    for (n = 0; n < corpus->num_docs; n++){
	phi[n] = malloc(sizeof(double) * NTOPICS);
    }

    // initialize model
    lda_suffstats* ss = NULL;
    mexPrintf("START IS SET TO:%d\n",start);
    if (start==0)
    {
        model = new_lda_model(corpus->num_terms, NTOPICS);
        ss = new_lda_suffstats(model);
        corpus_initialize_ss(ss, model, corpus);
        lda_mle(model, ss, 0);
        model->alpha = INITIAL_ALPHA;
    }
    else if (start==1)
    {
        model = new_lda_model(corpus->num_terms, NTOPICS);
        ss = new_lda_suffstats(model);
        random_initialize_ss(ss, model);
        lda_mle(model, ss, 0);
        model->alpha = INITIAL_ALPHA;
    }
    else
    {
        model = load_lda_model(start);
        ss = new_lda_suffstats(model);
    }
    model->nu=*nu;

    // run expectation maximization

    int i = 0;
    int k = 0;
    double* phi_s1=  malloc(sizeof(double) * NTOPICS);
    double* phi_s2=  malloc(sizeof(double) * NTOPICS);

  

    double likelihood, likelihood_old = 0, converged = 1;
    while (((converged < 0) || (converged > EM_CONVERGED) || (i <= 2)) && (i <= EM_MAX_ITER))
    {
        i++; mexPrintf("**** em iteration %d ****\n", i);
        likelihood = 0;
        zero_initialize_ss(ss, model);

        for(k=0;k<NTOPICS;k++){
           phi_s1[k]=0;
           phi_s2[k]=0;
        }



        // e-step

        for (d = 0; d < corpus->num_docs; d++)
        {
            if((corpus->docs[d]).length==0){
               continue;
            }
            if ((d % 1000) == 0) mexPrintf("document %d\n",d);
            likelihood += doc_e_step(&(corpus->docs[d]),
                                     var_gamma[d],
                                     phi[d],
                                     model,
                                     ss);

            for(k=0;k<NTOPICS;k++){
                 phi_s1[k]+=(double)corpus->docs[d].label*phi[d][k];     
                 phi_s2[k]+=1/(double)(corpus->docs[d].xz)*phi[d][k];
            }

        }


        // compute nu
        for(k=0;k<NTOPICS;k++){
          if(phi_s1[k]==0){
             phi_s1[k]=0.000001;
          } 
          model->nu[k]=log(phi_s1[k]/phi_s2[k]);
          if(model->nu[k]<0.000001 && phi_s1[k]<0.000001){
              phi_s1[k]=0.000001;
              model->nu[k]=log(phi_s1[k]/phi_s2[k]);   
          }
          //mexPrintf("NU: %f, PHI_SUM_1: %f, PHI_SUM_2: %f\n",model->nu[k],phi_s1[k],phi_s2[k]);
        }


        // m-step
        lda_mle(model, ss, ESTIMATE_ALPHA);

        // check for convergence

        converged = (likelihood_old - likelihood) / (likelihood_old);
        printf("CONVERGED: %10.10f\t%5.5e, %f,%f\n", likelihood, converged,converged,likelihood_old);

        if (converged < 0) VAR_MAX_ITER = VAR_MAX_ITER * 2;
        likelihood_old = likelihood;

        // output model and likelihood

     
    }

    for (d = 0; d < corpus->num_docs; d++)
    {
        if((corpus->docs[d]).length==0){
           continue;
        }

        if ((d % 100) == 0) mexPrintf("final e step document %d\n",d);
        likelihood += lda_inference(&(corpus->docs[d]), model, var_gamma[d], phi[d]);
    }
  
    mexPrintf("MODEL TOPICS: %d, MODEL TERMS:%d\n",model->num_terms,model->num_topics);
    *var_g=var_gamma;
    *phii=phi;
    return model;
}





corpus* init_data(int  *w, int *d, int *cs,int *lengths, int *counts,int nw,int nd,int nlength,int *cl)
{
 
    int n,i,z,doc,ind;
    corpus* c;

   
    c = (corpus*)malloc(sizeof(corpus));
    c->docs = 0;
    c->num_terms = 0;
    c->num_docs = 0;
  
    z=0;
    c->docs = (document*)malloc(sizeof(document)*(nd));
    
    for(i=0; i<nd;i++){
        //mexPrintf("DOCUMENT: %d, LENGTH: %d COUNT TOTAL:%d\n",i,lengths[i],counts[i]);
        c->docs[i].length = lengths[i];
	c->docs[i].total = counts[i];
	c->docs[i].words = malloc(sizeof(int)*lengths[i]);
	c->docs[i].counts = malloc(sizeof(int)*lengths[i]);
        c->docs[i].label = cl[i]; 
        c->docs[i].phi_sum = (double*)mxCalloc( NTOPICS , sizeof( double ));
        c->docs[i].xz=1;       
 
        ind=0;
        //go through counts and words 
        while(z<nlength && d[z]==i){
           //mexPrintf(" %d:%d ",w[z],cs[z]);
           c->docs[i].words[ind]=w[z];
           c->docs[i].counts[ind]= cs[z];              
           z++;
           ind++;
        }
        mexPrintf("\n");
    }
 
    c->num_docs = nd;
    c->num_terms = nw;
    return(c);
}






void mexFunction(int nlhs, mxArray *plhs[], int nrhs,
                 const mxArray *prhs[])
{
  double *srwp, *srdp, *probs, *Z, *WS, *DS, *ZIN, *CS, *LABELS,*nu;
  double ALPHA,BETA;
  int *irwp, *jcwp, *irdp, *jcdp,*lengths,*counts;
  int *z,*d,*w,*c, *order, *wp, *dp, *ztot, *cl;
  int W,T,D, nzmax, nzmaxwp, nzmaxdp, nlength,start;
  int i,j,n,nt,wi,di, startcond;
  corpus* corp;
  double **var_gamma,**phi;
  lda_model *model;
  
  /* Check for proper number of arguments. */
  if (nrhs < 10) {
    mexErrMsgTxt("At least 10 input arguments required");
  } else if (nlhs < 3) {
    
  }
  

  
  /* process the input arguments */
  if (mxIsDouble( prhs[ 0 ] ) != 1) mexErrMsgTxt("WS input vector must be a double precision matrix");
  if (mxIsDouble( prhs[ 1 ] ) != 1) mexErrMsgTxt("DS input vector must be a double precision matrix");
  if (mxIsDouble( prhs[ 2 ] ) != 1) mexErrMsgTxt("CS input vector must be a double precision matrix"); 

 
  // pointer to word indices
  WS = mxGetPr( prhs[ 0 ] );
     
  // pointer to document indices
  DS = mxGetPr( prhs[ 1 ] );
  
   // pointer to document indices
  CS = mxGetPr( prhs[ 2 ] );

  // get the number of tokens
  nlength = mxGetM( prhs[ 0 ] ) * mxGetN( prhs[ 0 ] );
  
  
  if (nlength == 0) mexErrMsgTxt("WS vector is empty"); 
  if (nlength != ( mxGetM( prhs[ 1 ] ) * mxGetN( prhs[ 1 ] ))) mexErrMsgTxt("WS and DS vectors should have same number of entries");
  if (nlength != ( mxGetM( prhs[ 2 ] ) * mxGetN( prhs[ 2 ] ))) mexErrMsgTxt("WS and DS vectors should have same number of entries");  

  NTOPICS    = (int) mxGetScalar(prhs[3]);
  if (NTOPICS<=0) mexErrMsgTxt("Number of topics must be greater than zero");
  
  VAR_MAX_ITER     = (int) mxGetScalar(prhs[4]);
  if (VAR_MAX_ITER <0) mexErrMsgTxt("Number of iterations for variational inf must be positive");
  
  EM_MAX_ITER    = (int) mxGetScalar(prhs[5]);
  if (EM_MAX_ITER<0) mexErrMsgTxt("Number of iterations for em  inf must be positive");


  VAR_CONVERGED= (double) mxGetScalar(prhs[6]);
  if (VAR_CONVERGED<=0) mexErrMsgTxt("VAR_CONVERGED must be greater than zero");
  
  EM_CONVERGED = (double) mxGetScalar(prhs[7]);
  if (EM_CONVERGED<=0) mexErrMsgTxt("EM_CONVERGED must be greater than zero");


  INITIAL_ALPHA = (double) mxGetScalar(prhs[8]);
  if (INITIAL_ALPHA<=0) mexErrMsgTxt("initial ALPHA must be greater than zero");
  
  start    = (int) mxGetScalar(prhs[9]);

  //read in document labels
  LABELS=mxGetPr( prhs[ 10] );

  LAP=(double) mxGetScalar(prhs[ 10]);

 
  
/* allocate memory */
  z  = (int *) mxCalloc( nlength , sizeof( int ));
  

  d  = (int *) mxCalloc( nlength , sizeof( int ));
  w  = (int *) mxCalloc( nlength , sizeof( int ));
  c  = (int *) mxCalloc( nlength , sizeof( int ));
 
  

  order  = (int *) mxCalloc( nlength , sizeof( int ));  
  ztot  = (int *) mxCalloc( NTOPICS , sizeof( int ));
  probs  = (double *) mxCalloc( NTOPICS , sizeof( double ));  
  nu= (double *) mxCalloc( NTOPICS , sizeof( double ));

  // copy over the word and document indices into internal format
  for (i=0; i<nlength; i++) {
     w[ i ] = (int) WS[ i ] - 1;
     d[ i ] = (int) DS[ i ] - 1;
     c[ i ] = (int) CS[ i ];
  }
  
  n = nlength;
  
  W = 0;
  D = 0;
  for (i=0; i<n; i++) {
     if (w[ i ] > W) W = w[ i ];
     if (d[ i ] > D) D = d[ i ];
  }
  W = W + 1;
  D = D + 1;
  

  lengths = (int *) mxCalloc( D , sizeof( int ));
  counts = (int *) mxCalloc( D , sizeof( int ));
  cl=(int *) mxCalloc( D , sizeof( int ));


  wp  = (int *) mxCalloc( NTOPICS*W , sizeof( int ));
  dp  = (int *) mxCalloc( NTOPICS*D , sizeof( int ));
   
  

  for(i=0;i<nlength;i++){
       lengths[d[i]]++;
       counts[d[i]]+=c[i];
  }
  
  mexPrintf("INIT LABELS.. \n");
  for(i=0;i<D;i++){
     cl[i]=(int)LABELS[i];
  }  

  mexPrintf("INIT ETA...\n");

  for(i=0;i<NTOPICS;i++){
    nu[i]=-1.0/(double)NTOPICS;
  }


  

  mexPrintf("INIT DATA...\n");
  ESTIMATE_ALPHA = 1;
  //initialize document corpus data structure
  corp=init_data(w, d, c, lengths, counts,W,D,nlength,cl);
  
  //run em 
  mexPrintf("RUN EM...\n"); 
  model=run_em_mat(start, corp, &var_gamma, &phi, &nu);
 
 
  //now return values
  // ALHPA, NUM OF TOPICS, NUM OF TERMS, LOG_PROBS, GAMMA, WORD ASSIGNMENT
  
  //ALPHA
  plhs[ 0 ] = mxCreateDoubleMatrix( 1,1 , mxREAL );
  Z = mxGetPr( plhs[ 0 ] );
  Z[0]=(double) model->alpha;

  
  //NUM OF TOPICS
   plhs[ 1 ] = mxCreateDoubleMatrix( 1,1 , mxREAL );
   Z = mxGetPr( plhs[ 1 ] );
   Z[0]=(double) NTOPICS;


   //NUM OF TERMS
   plhs[ 2 ] = mxCreateDoubleMatrix( 1,1 , mxREAL );
   Z = mxGetPr( plhs[ 2 ] );
   Z[0]=(double) W;


   //LOG PROBS
   plhs[ 3 ] = mxCreateDoubleMatrix( model->num_terms, model->num_topics,mxREAL );
   Z = mxGetPr( plhs[ 3 ] );
   
   
 
   for (i = 0; i < model->num_topics; i++)
    {
     for (j = 0; j < model->num_terms; j++)
      {
   	  Z[j+i*model->num_terms]=model->log_prob_w[i][j];
      }
    }
   


   //GAMMA MATRIX
   plhs[ 4 ] = mxCreateDoubleMatrix( D, NTOPICS,mxREAL );
   Z = mxGetPr( plhs[ 4 ] );
   for (i = 0; i < D; i++)
    {
     for (j = 0; j < NTOPICS; j++)
      {
          Z[i+j*D]=var_gamma[i][j];
      }
    }


   //NU VECTOR
   plhs[ 5 ] = mxCreateDoubleMatrix( NTOPICS,1,mxREAL );
   Z = mxGetPr( plhs[ 5 ] );
   for (j = 0; j < NTOPICS; j++)
   {
      Z[j]=nu[j];
   }
   

     
}
