%INPUTS
%  model - a struct specifying model parameters. It needs to contain the following fields
%        model.ntopics        -  number of topics (mixture components)
%        model.initial_alpha  -  initial parameter for the dirichlet distribution (same entry assumed for vector)
%        model.lap            -  smoothing parameter, to ensure that some mixture weights do not go to zero
%        model.var_converged  -  stopping condition for variational inference
%        model.var_max_iter   -  maximum number of iterations for variational inference (sub loop in em algorithm)
%        model.em_converged   -  stopping condition for EM algorithm
%        model.em_max_iter    -  maximum number of iterations for em algorithm
%
%
% data - a MATLAB struct specifying a data set with the following fields:
%       data.name    -   name of the data set    
%       data.counts  -   A matrix with three columns. The first column is assumed to be the document
%              id. Document ids have to start with a one and be consecutive.
%
%              The second column indicates the word id. Word id is an index into the vocabulary.
%              The lowest index is assumed to be one.
%
%              The third column indicates how many times a given word occurs in a given document.
%
%              Examlple [1  34  8
%                        1  3   2
%                        1  53  4
%                        1  45  12
%                        2  2   4
%                        2  9   2
%                        2  7   5
%                        2  2   8
%                        3  4   2
%                        3  26  6
%                        3  65  9]
%
%     data.words - cell array containing the words in the vocabulary. The indecies of the words should be the same as used in 
%                  the counts matrix
%     data.labels - binary vector indicating label of each document. The length of this vector should be equal to number of 
%                  documents.
%
%OUTPUTS:
%  model -  the trained model represented by a MATLAB struct. The following fields are learned and added to the struct
%        model.alpha - learned alpha parameter
%        model.wp    - Assuming that there are N words and K topics. This is a NxK matrix indicating
%                      the logliklihood for each word within the respective topics
%        model.gamma - If there are D documents and K topics. This is a DxK matrix. These are variational parameters
%                      which were assumed to generate documents. Once normalized these can be considered as approximations
%                      for mixture weights.
%        model.nu    - parameters for the logistic regression part of the model
%        model.terms - number of words in vocabulary
%

function[model]=train(model,data)

labels=data.labels;
counts=data.counts;

doc=counts(:,1);
wrd=counts(:,2);
cnt=counts(:,3);


[alpha,ntopics,terms,WP,GAMMA,NU]=fdldaestimate(wrd,doc,cnt,model.ntopics,model.var_max_iter,model.em_max_iter,...
                                  model.var_converged, model.em_converged,model.initial_alpha,0,labels,model.lap);

model.alpha=alpha;
model.terms=terms;
model.wp=WP;
model.gamma=GAMMA;
model.nu=NU;



